// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:flutter/foundation.dart';
import 'package:wyatt_authentication_bloc/src/form/form_entry.dart';
import 'package:wyatt_authentication_bloc/src/form/form_input.dart';
import 'package:wyatt_authentication_bloc/src/models/validation_error.dart';

@immutable
class FormData {
  const FormData(this._entries);

  FormData.empty() : this(<FormEntry>[]);

  final List<FormEntry> _entries;

  List<FormEntry> get entries => _entries;

  List<FormInput<T, ValidationError>> inputs<T>() {
    return _entries
        .map((FormEntry entry) => entry.input as FormInput<T, ValidationError>)
        .toList();
  }

  FormInput<T, ValidationError> input<T>(String field) {
    if (contains(field)) {
      return _entries
          .firstWhere((FormEntry entry) => entry.field == field)
          .input as FormInput<T, ValidationError>;
    } else {
      throw Exception('Field $field does not exist in form');
    }
  }

  bool contains(String field) {
    return _entries.any((FormEntry entry) => entry.field == field);
  }

  FormData intersection(FormData other) {
    final List<FormEntry> entries = <FormEntry>[];

    for (final FormEntry entry in _entries) {
      if (other.contains(entry.field)) {
        entries.add(entry);
      }
    }

    return FormData(entries);
  }

  FormData difference(FormData other) {
    final List<FormEntry> entries = <FormEntry>[];

    for (final FormEntry otherEntry in other._entries) {
      if (!contains(otherEntry.field)) {
        entries.add(otherEntry);
      }
    }

    for (final FormEntry entry in _entries) {
      if (!other.contains(entry.field)) {
        entries.add(entry);
      }
    }

    return FormData(entries);
  }

  FormData union(FormData other) {
    final List<FormEntry> entries = <FormEntry>[];

    for (final FormEntry entry in _entries) {
      entries.add(entry);
    }

    for (final FormEntry otherEntry in other._entries) {
      if (!contains(otherEntry.field)) {
        entries.add(otherEntry);
      }
    }

    return FormData(entries);
  }

  void update(String field, FormInput input) {
    if (contains(field)) {
      final index = _entries.indexOf(
        _entries.firstWhere((FormEntry entry) => entry.field == field),
      );
      _entries[index] = _entries[index].copyWith(input: input);
    }
  }

  FormData clone() {
    return FormData(
      _entries
          .map((FormEntry entry) => entry.clone())
          .toList(),
    );
  }

  Map<String, dynamic> toMap() {
    final map = <String, dynamic>{};
    for (final entry in _entries) {
      if (entry.export) {
        map[entry.fieldName ?? entry.field] = entry.input.value;
      }
    }
    return map;
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;

    return other is FormData && listEquals(other._entries, _entries);
  }

  @override
  int get hashCode => _entries.hashCode;

  @override
  String toString() => 'FormData(entries: $_entries)';
}
