// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:firebase_auth/firebase_auth.dart';
import 'package:wyatt_authentication_bloc/src/models/exceptions/exceptions_firebase.dart';
import 'package:wyatt_authentication_bloc/src/models/user/user_firebase.dart';
import 'package:wyatt_authentication_bloc/src/models/user/user_interface.dart';
import 'package:wyatt_authentication_bloc/src/repositories/authentication_repository_interface.dart';

class AuthenticationRepositoryFirebase
    implements AuthenticationRepositoryInterface {
  final FirebaseAuth _firebaseAuth;

  UserFirebase _userCache = const UserFirebase.empty();

  AuthenticationRepositoryFirebase({FirebaseAuth? firebaseAuth})
      : _firebaseAuth = firebaseAuth ?? FirebaseAuth.instance;

  @override
  Stream<UserInterface> get user {
    return _firebaseAuth.userChanges().map((User? firebaseUser) {
      final UserFirebase user = firebaseUser == null
          ? const UserFirebase.empty()
          : firebaseUser.model;
      _userCache = user;
      return user;
    });
  }

  @override
  UserInterface get currentUser {
    return _userCache;
  }

  @override
  Future<void> applyActionCode(String code) async {
    try {
      await _firebaseAuth.applyActionCode(code);
    } on FirebaseAuthException catch (e) {
      throw ApplyActionCodeFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw ApplyActionCodeFailureFirebase();
    }
  }

  @override
  Future<String?> signUp({
    required String email,
    required String password,
  }) async {
    try {
      final creds = await _firebaseAuth.createUserWithEmailAndPassword(
        email: email,
        password: password,
      );
      return creds.user?.uid;
    } on FirebaseAuthException catch (e) {
      throw SignUpWithEmailAndPasswordFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw SignUpWithEmailAndPasswordFailureFirebase();
    }
  }

  @override
  Future<List<String>> fetchSignInMethodsForEmail({
    required String email,
  }) async {
    try {
      return await _firebaseAuth.fetchSignInMethodsForEmail(email);
    } on FirebaseAuthException catch (e) {
      throw FetchSignInMethodsForEmailFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw FetchSignInMethodsForEmailFailureFirebase();
    }
  }

  @override
  Future<void> signInAnonymously() async {
    try {
      await _firebaseAuth.signInAnonymously();
    } on FirebaseAuthException catch (e) {
      throw SignInAnonymouslyFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw SignInAnonymouslyFailureFirebase();
    }
  }

  @override
  Future<void> signInWithGoogle() {
    // TODO(hpcl): implement signInWithGoogle
    throw UnimplementedError();
  }

  @override
  Future<void> signInWithEmailLink(String email, String emailLink) async {
    try {
      await _firebaseAuth.signInWithEmailLink(
        email: email,
        emailLink: emailLink,
      );
    } on FirebaseAuthException catch (e) {
      throw SignInWithEmailLinkFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw SignInWithEmailLinkFailureFirebase();
    }
  }

  @override
  Future<void> signInWithEmailAndPassword({
    required String email,
    required String password,
  }) async {
    try {
      await _firebaseAuth.signInWithEmailAndPassword(
        email: email,
        password: password,
      );
    } on FirebaseAuthException catch (e) {
      throw SignInWithEmailAndPasswordFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw SignInWithEmailAndPasswordFailureFirebase();
    }
  }

  @override
  Future<void> sendEmailVerification() async {
    try {
      await _userCache.inner!.sendEmailVerification();
    } catch (e) {
      throw SendEmailVerificationFailureFirebase();
    }
  }

  @override
  Future<void> sendPasswordResetEmail({required String email}) async {
    try {
      await _firebaseAuth.sendPasswordResetEmail(email: email);
    } on FirebaseAuthException catch (e) {
      throw SendPasswordResetEmailFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw SendPasswordResetEmailFailureFirebase();
    }
  }

  @override
  Future<void> sendSignInLinkEmail({required String email}) async {
    try {
      // TODO(hpcl): implement sendSignInLinkEmail
    } on FirebaseAuthException catch (e) {
      throw SendSignInLinkEmailFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw SendSignInLinkEmailFailureFirebase();
    }
  }

  @override
  Future<void> confirmPasswordReset({
    required String code,
    required String newPassword,
  }) async {
    try {
      await _firebaseAuth.confirmPasswordReset(
        code: code,
        newPassword: newPassword,
      );
    } on FirebaseAuthException catch (e) {
      throw ConfirmPasswordResetFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw ConfirmPasswordResetFailureFirebase();
    }
    throw UnimplementedError();
  }

  @override
  Future<void> verifyPasswordResetCode({required String code}) async {
    try {
      await _firebaseAuth.verifyPasswordResetCode(code);
    } on FirebaseAuthException catch (e) {
      throw VerifyPasswordResetCodeFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw VerifyPasswordResetCodeFailureFirebase();
    }
  }

  @override
  Future<void> signOut() async {
    try {
      await Future.wait([
        _firebaseAuth.signOut(),
      ]);
      _userCache = const UserFirebase.empty();
    } catch (_) {
      throw SignOutFailureFirebase();
    }
  }

  @override
  Future<void> refresh() async {
    try {
      await _userCache.inner!.reload();
    } on FirebaseAuthException catch (e) {
      throw RefreshFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw RefreshFailureFirebase();
    }
  }
}

extension on User {
  UserFirebase get model {
    return UserFirebase(this);
  }
}
