// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:wyatt_authentication_bloc/src/models/exceptions/exceptions_interface.dart';
import 'package:wyatt_authentication_bloc/src/models/user/user_interface.dart';

/// {@template authentication_repository}
/// Repository which manages user authentication.
/// {@endtemplate}
abstract class AuthenticationRepositoryInterface {
  /// Stream of [UserInterface] which will emit the current user when
  /// the authentication state changes.
  ///
  /// Emits [UserInterface.empty] if the user is not authenticated.
  Stream<UserInterface> get user;

  /// Returns the current cached account.
  /// Defaults to [UserInterface.empty] if there is no cached user.
  UserInterface get currentUser;

  /// Applies action code
  ///
  /// [code] - The action code sent to the user's email address.
  /// Throw [ApplyActionCodeFailureInterface] if an exception occurs.
  Future<void> applyActionCode(String code);

  /// Creates a new user with the provided [email] and [password].
  ///
  /// Returns the newly created user's unique identifier.
  ///
  /// Throws a [SignUpWithEmailAndPasswordFailureInterface] if
  /// an exception occurs.
  Future<String?> signUp({required String email, required String password});

  /// Fetches sign in methods for [email].
  ///
  /// Throws a [FetchSignInMethodsForEmailFailureInterface] if
  /// an exception occurs.
  Future<List<String>> fetchSignInMethodsForEmail({required String email});

  /// Sign in anonimously.
  ///
  /// Throws a [SignInAnonymouslyFailureInterface] if an exception occurs.
  Future<void> signInAnonymously();

  /// Starts the Sign In with Google Flow.
  ///
  /// Throws a [SignInWithGoogleFailureInterface] if an exception occurs.
  Future<void> signInWithGoogle();

  /// Signs in using an email address and email sign-in link.
  ///
  /// Throws a [SignInWithEmailLinkFailureInterface] if an exception occurs.
  Future<void> signInWithEmailLink(
    String email,
    String emailLink,
  );

  /// Signs in with the provided [email] and [password].
  ///
  /// Throws a [SignInWithEmailAndPasswordFailureInterface] if
  /// an exception occurs.
  Future<void> signInWithEmailAndPassword({
    required String email,
    required String password,
  });

  /// Sends verification email to the provided [user].
  ///
  /// Throws a [SendEmailVerificationFailureInterface] if an exception occurs.
  Future<void> sendEmailVerification();

  /// Sends a password reset email to the provided [email].
  ///
  /// Throws a [SendPasswordResetEmailFailureInterface] if an exception occurs.
  Future<void> sendPasswordResetEmail({required String email});

  /// Sends link to login.
  ///
  /// Throws a [SendSignInLinkEmailFailureInterface] if an exception occurs.
  Future<void> sendSignInLinkEmail({required String email});

  /// Confirms the password reset with the provided [newPassword] and [code].
  ///
  /// Throws a [ConfirmPasswordResetFailureInterface] if an exception occurs.
  Future<void> confirmPasswordReset({
    required String code,
    required String newPassword,
  });

  /// Verify password reset code.
  ///
  /// Throws a [VerifyPasswordResetCodeFailureInterface] if an exception occurs.
  Future<void> verifyPasswordResetCode({required String code});

  /// Signs out the current user which will emit
  /// [UserInterface.empty] from the [user] Stream.
  Future<void> signOut();

  /// Refreshes the current user.
  ///
  /// Throws a [RefreshFailureInterface] if an exception occurs.
  Future<void> refresh();
}
