// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:bloc/bloc.dart';
import 'package:flutter/foundation.dart';
import 'package:wyatt_authentication_bloc/src/authentication/cubit/authentication_cubit.dart';
import 'package:wyatt_authentication_bloc/src/form/form_data.dart';
import 'package:wyatt_authentication_bloc/src/form/form_input.dart';
import 'package:wyatt_authentication_bloc/src/form/form_status.dart';
import 'package:wyatt_authentication_bloc/src/form/form_validator.dart';
import 'package:wyatt_authentication_bloc/src/models/email.dart';
import 'package:wyatt_authentication_bloc/src/models/exceptions/exceptions_interface.dart';
import 'package:wyatt_authentication_bloc/src/models/operations.dart';
import 'package:wyatt_authentication_bloc/src/models/password.dart';
import 'package:wyatt_authentication_bloc/src/models/validation_error.dart';
import 'package:wyatt_authentication_bloc/src/repositories/authentication_repository_interface.dart';

part 'sign_up_state.dart';

class SignUpCubit extends Cubit<SignUpState> {
  final AuthenticationRepositoryInterface _authenticationRepository;
  final AuthenticationCubit _authenticationCubit;

  final Future Function(SignUpState state, String? uid)? _onSignUpSuccess;

  SignUpCubit({
    required AuthenticationRepositoryInterface authenticationRepository,
    required AuthenticationCubit authenticationCubit,
    required FormData entries,
    Future Function(SignUpState state, String? uid)? onSignUpSuccess,
  })  : _authenticationRepository = authenticationRepository,
        _authenticationCubit = authenticationCubit,
        _onSignUpSuccess = onSignUpSuccess,
        super(SignUpState(data: entries));

  void emailChanged(String value) {
    final Email email = Email.dirty(value);

    final List<FormInput<dynamic, ValidationError>> inputsToValidate = [
      email,
      state.password,
      ...state.data.inputs<dynamic>(),
    ];

    emit(
      state.copyWith(
        email: email,
        status: FormValidator.validate(inputsToValidate),
      ),
    );
  }

  void passwordChanged(String value) {
    final Password password = Password.dirty(value);

    final List<FormInput<dynamic, ValidationError>> inputsToValidate = [
      state.email,
      password,
      ...state.data.inputs<dynamic>(),
    ];

    emit(
      state.copyWith(
        password: password,
        status: FormValidator.validate(inputsToValidate),
      ),
    );
  }

  void dataChanged<T>(String field, FormInput dirtyValue) {
    final _form = state.data.clone();

    if (_form.contains(field)) {
      _form.update(field, dirtyValue);
    } else {
      throw Exception('Form field $field not found');
    }

    emit(
      state.copyWith(
        data: _form,
        status: FormValidator.validate(
          [state.email, state.password, ..._form.inputs<dynamic>()],
        ),
      ),
    );
  }

  void updateFormData(
    FormData data, {
    SetOperation operation = SetOperation.replace,
  }) {
    FormData _form = data;

    switch (operation) {
      case SetOperation.replace:
        _form = data;
        break;
      case SetOperation.difference:
        _form = state.data.difference(data);
        break;
      case SetOperation.intersection:
        _form = state.data.intersection(data);
        break;
      case SetOperation.union:
        _form = state.data.union(data);
        break;
    }

    emit(
      state.copyWith(
        data: _form,
        status: FormValidator.validate(
          [state.email, state.password, ..._form.inputs<dynamic>()],
        ),
      ),
    );
  }

  Future<void> signUpFormSubmitted() async {
    if (!state.status.isValidated) return;
    emit(state.copyWith(status: FormStatus.submissionInProgress));
    try {
      final uid = await _authenticationRepository.signUp(
        email: state.email.value,
        password: state.password.value,
      );
      await _onSignUpSuccess?.call(state, uid);
      if (_authenticationCubit.start()) {
        emit(state.copyWith(status: FormStatus.submissionSuccess));
      } else {
        emit(
          state.copyWith(
            errorMessage: 'Failed to start authentication cubit',
            status: FormStatus.submissionFailure,
          ),
        );
      }
    } on SignUpWithEmailAndPasswordFailureInterface catch (e) {
      emit(
        state.copyWith(
          errorMessage: e.message,
          status: FormStatus.submissionFailure,
        ),
      );
    } catch (e) {
      debugPrint(e.toString());
      emit(state.copyWith(status: FormStatus.submissionFailure));
    }
  }
}
